from pydantic import NameEmail

__all__ = ["dario", "jakob", "wei_hsin", "samuel_saraiva"]


dario = NameEmail(name="Dario Cambiè", email="2422614+dcambie@users.noreply.github.com")
jakob = NameEmail(name="Jakob Wolf", email="Jakob.Wolf@mpikg.mpg.de")
wei_hsin = NameEmail(name="Wei-Hsin Hsu", email="Wei-hsin.Hsu@mpikg.mpg.de")
miguel = NameEmail(name="Miguel Garcia", email="miguel.garcia@mpikg.mpg.de")
samuel_saraiva = NameEmail(name="Samuel Saraiva", email="Samuel.Saraiva@mpikg.mpg.de")

